# Hye-Shik Chang <16 Feb 2002>
# $Id: qwerty2bul.py,v 1.1.1.1 2002/02/16 00:51:36 perky Exp $

import codecs
from korean.hangul import Moeum, Jaeum, Chosung, Jungsung, Jongsung
from korean.hangul import ishangul, join, split, isJaeum, isMoeum

codekeymap = {
        Jaeum.G: 'r',        Jaeum.GG: 'R',       Jaeum.GS: 'rt',
        Jaeum.N: 's',        Jaeum.NJ:'sw',       Jaeum.NH: 'sg',       Jaeum.D: 'e',
        Jaeum.DD:'E',        Jaeum.R: 'f',        Jaeum.RG: 'fr',       Jaeum.RM: 'fa',
        Jaeum.RB:'fq',       Jaeum.RS:'ft',       Jaeum.RT: 'fx',       Jaeum.RP: 'fv',
        Jaeum.RH:'fg',       Jaeum.M: 'a',        Jaeum.B:  'q',        Jaeum.BB: 'Q',
        Jaeum.BS:'qt',       Jaeum.S: 't',        Jaeum.SS: 'T',        Jaeum.A:  'd',
        Jaeum.J: 'w',        Jaeum.JJ:'W',        Jaeum.CH: 'c',        Jaeum.K:  'z',
        Jaeum.T: 'x',        Jaeum.P: 'v',        Jaeum.H:  'g',

        Moeum.A: 'k',        Moeum.AE:'o',        Moeum.YA: 'i',        Moeum.YAE:'O',
        Moeum.EO:'j',        Moeum.E: 'p',        Moeum.YEO:'u',        Moeum.YE: 'P',
        Moeum.O: 'h',        Moeum.WA:'hk',       Moeum.WAE:'ho',       Moeum.WOE: 'hl',
        Moeum.YO:'y',        Moeum.OO:'n',        Moeum.WO: 'nj',       Moeum.WE: 'np',
        Moeum.WI:'nl',       Moeum.YU:'b',        Moeum.EU: 'm',        Moeum.EUI: 'ml',
        Moeum.I: 'l',

        u'': '',
}

keycodemap = {}
for k, v in codekeymap.items():
        keycodemap[v] = k
        keycodemap.setdefault(v.upper(), k)
keycodes = ''.join(keycodemap.keys())


class Automata_Hangul2:
    
    # must Unicode in / Unicode out

    def __init__(self):
        self.clear()

    def pushcomp(self):
        if self.chosung and not self.jungsung:
            self.word_valid = 0
        self.word_comp.append(join([self.chosung, self.jungsung, self.jongsung]))
        self.clearcomp()

    def clearcomp(self):
        self.chosung = u''
        self.jungsung = u''
        self.jongsung = u''

    def clear(self):
        self.buff = ['']
        self.word_raw = []
        self.word_comp = []
        self.word_valid = 1
        self.clearcomp()

    def convert(self, s):
        self.clear()

        map(self.feed, s)
        self.finalize()

        return u''.join(self.buff)
    
    def finalize(self):
        if self.chosung or self.jungsung or self.jongsung:
            self.pushcomp()
        if self.word_raw or self.word_comp:
            if self.word_valid:
                self.buff.append(u''.join(self.word_comp))
            else:
                self.word_valid = 1
                self.buff.append(u''.join(self.word_raw))
            
            self.word_raw, self.word_comp = [], []

    def feed(self, c):
        self.word_raw.append(c)
        if c in keycodes:
            code = keycodemap[c]
            if isJaeum(code):
                if not self.chosung: # chosung O
                    if self.jungsung or self.jongsung:
                        self.word_valid = 0
                    else:
                        self.chosung = code
                elif not self.jungsung: # chosung O  jungsung X
                    if self.jongsung:
                        self.word_valid = 0
                    else:
                        self.pushcomp()
                        self.chosung = code
                elif not self.jongsung: # chosung O  jungsung O  jongsung X
                    if code not in Jongsung:
                        self.pushcomp()
                        self.chosung = code
                    else:
                        self.jongsung = code
                else: # full
                    trymul = codekeymap[self.jongsung] + c
                    if keycodemap.has_key(trymul): # can be multi jongsung
                        self.jongsung = keycodemap[trymul]
                    else:
                        self.pushcomp()
                        self.chosung = code
            else: # MOEUM...
                if not self.jongsung:
                    if not self.jungsung: # jungsung X  jongsung X
                        self.jungsung = code
                    else: # jungsung O  jongsung X
                        trymul = codekeymap[self.jungsung] + c
                        if keycodemap.has_key(trymul): # can be multi jungsung
                            self.jungsung = keycodemap[trymul]
                        else:
                            self.pushcomp()
                            self.jungsung = code
                else: # jongsung O
                    if len(codekeymap[self.jongsung]) > 1:
                        ojong = keycodemap[codekeymap[self.jongsung][:-1]]
                        ncho  = keycodemap[codekeymap[self.jongsung][-1]]
                        self.jongsung = ojong
                        self.pushcomp()
                        self.chosung = ncho
                        self.jungsung = code
                    else:
                        njong = self.jongsung
                        self.jongsung = u''
                        self.pushcomp()
                        self.chosung = njong
                        self.jungsung = code
        else: # non key code
            self.finalize()
            self.buff.append(c)


class Codec(codecs.Codec):

    BASECODEC = 'korean.cp949' # fallback codec of decoder

    # Unicode to key stroke
    def encode(self, data, errors='strict',
               supported_errors=('strict', 'ignore', 'replace')):
        if errors not in supported_errors:
            raise UnicodeError, "unknown error handling"

        r = []
        for c in data:
            if c <= u'\u0080':
                r.append(c.encode('ascii'))
            elif not ishangul(c):
                r.append(c.encode(self.BASECODEC, errors=errors))
            else:
                for k in split(c):
                    r.append(codekeymap[k])

        r = ''.join(r)
        return (r, len(r))

    # key stroke to Unicode
    def decode(self, data, errors='strict',
               supported_errors=('strict', 'ignore', 'replace')):
        if errors not in supported_errors:
            raise UnicodeError, "unknown error handling"

        s = unicode(data, self.BASECODEC, errors=errors)
        am = Automata_Hangul2()
        r = am.convert(s)
        return (r, len(r))

class StreamWriter(Codec, codecs.StreamWriter):
    pass

class StreamReader(Codec, codecs.StreamReader):
    pass

def getregentry(*args):
    return (Codec().encode, Codec().decode, StreamReader, StreamWriter)

def getaliases():
    return ('2bul', '2bulsik',)



# A codec for ISO-2022-KR [RFC1557]
# by Hye-Shik Chang <17 February 2002>
#
# StreamReader was adopted from Tamito KAJIYAMA's iso-2022-jp codec.
#
# $Id: iso_2022_kr.py,v 1.3 2002/02/16 11:34:25 perky Exp $

import codecs
from korean.mappings import ksc5601_hangul
encmap_hangul = ksc5601_hangul.encoding_map
decmap_hangul = ksc5601_hangul.decoding_map
encmap_ideo, decmap_ideo = {}, {}
encmap_misc, decmap_misc = {}, {}

US_ASCII      = 1
KSC5601_1987  = 2

CHARSETS = {
    "\033(B": US_ASCII,
    "\033$)C": KSC5601_1987,
}
SI = '\x0f'
SO = '\x0e'
ESC = '\033'

DESIGNATIONS = {}
for k, v in CHARSETS.items():
    DESIGNATIONS[v] = k

class Codec(codecs.Codec):
    # Unicode to character buffer
    def encode(self, data, errors='strict',
               supported_errors=('strict', 'ignore', 'replace')):
        global encmap_ideo, encmap_misc

        if errors not in supported_errors:
            raise UnicodeError, "unknown error handling"
        buffer = []
        charset = US_ASCII
        shiftstate = 0
        for c in data:
            if c in ('\n', '\r'):
                shiftstate = 0

            if c < u'\u0080':
                new_charset = US_ASCII
                s = c.encode("ascii", errors)
            elif encmap_hangul.has_key(c):
                new_charset = KSC5601_1987
                s = encmap_hangul[c]
            else:
                if not encmap_misc:
                    from korean.mappings import ksc5601_misc
                    encmap_misc = ksc5601_misc.encoding_map
                if encmap_misc.has_key(c):
                    new_charset = KSC5601_1987
                    s = encmap_misc[c]
                else:
                    if not encmap_ideo:
                        from korean.mappings import ksc5601_ideograph
                        encmap_ideo = ksc5601_ideograph.encoding_map
                    if encmap_ideo.has_key(c):
                        new_charset = KSC5601_1987
                        s = encmap_ideo[c]
                    elif errors == 'replace':
                        new_charset = KSC5601_1987
                        s = '\xa0\xa0'
                    elif errors == 'strict':
                        raise UnicodeError, "cannot map \\u%04x to ISO-2022-KR" % ord(c)
                    else:
                        continue

            if charset != new_charset:
                charset = new_charset
                if charset == US_ASCII:
                    if shiftstate:
                        buffer.append(SI)
                        shiftstate = 0
                else:
                    buffer.append(DESIGNATIONS[charset])
                    if not shiftstate:
                        buffer.append(SO)
                        shiftstate = 1

            if charset == KSC5601_1987 and shiftstate:
                s = chr(ord(s[0])&0x7F) + chr(ord(s[1])&0x7F)
            buffer.append(s)
        if shiftstate:
            buffer.append(SI)
            #buffer.append(DESIGNATIONS[US_ASCII])
        return (''.join(buffer), len(data))

    # character buffer to Unicode
    def decode(self, data, errors='strict',
               supported_errors=('strict', 'ignore', 'replace')):
        global decmap_ideo, decmap_misc
        
        if errors not in supported_errors:
            raise UnicodeError, "unknown error handling"
        buffer = []
        data = str(data) # character buffer compatible object
        size = len(data)
        charset = US_ASCII
        shiftstate = 0
        escstart = -1
        p = 0
        while p < size:
            if data[p] in ('\n', '\r'):
                shiftstate = 0

            if escstart >= 0:
                if data[p].isalpha():
                    escstr = data[escstart:p+1]
                    if CHARSETS.has_key(escstr):
                        charset = CHARSETS[escstr]
                    elif errors == 'strict':
                        raise UnicodeError, "unsupported charset found"
                    escstart = -1
                p += 1
            elif data[p] == SO:
                shiftstate = 1
                p += 1
            elif data[p] == SI:
                shiftstate = 0
                p += 1
            elif data[p] == ESC:
                escstart = p
                p += 1
            else:
                if not shiftstate and (
                        charset == US_ASCII or data[p] < '\x80'): # ascii
                    buffer.append(unicode(data[p], "ascii", errors))
                    p += 1
                else:
                    c = data[p:p+2]
                    if len(c) == 2:
                        c = chr(ord(c[0])|0x80) + chr(ord(c[1])|0x80)
                        if decmap_hangul.has_key(c):
                            buffer.append(decmap_hangul[c])
                            p += 2
                            continue

                        if not decmap_misc:
                            from korean.mappings import ksc5601_misc
                            decmap_misc = ksc5601_misc.decoding_map
                        if decmap_misc.has_key(c):
                            buffer.append(decmap_misc[c])
                            p += 2
                            continue

                        if not decmap_ideo:
                            from korean.mappings import ksc5601_ideograph
                            decmap_ideo = ksc5601_ideograph.decoding_map
                        if decmap_ideo.has_key(c):
                            buffer.append(decmap_ideo[c])
                            p += 2
                            continue

                    if errors == 'replace':
                        buffer.append(u'\uFFFD') # REPLACEMENT CHARACTER
                    elif errors == 'strict':
                        raise UnicodeError, "unexpected byte 0x%02x found" % ord(data[p])
                    p += 1
        
        return (u''.join(buffer), len(data))

class StreamWriter(Codec, codecs.StreamWriter):
    pass

class StreamReader(Codec, codecs.StreamReader):
    def __init__(self, stream, errors='strict'):
        codecs.StreamReader.__init__(self, stream, errors)
        self.data = ''
        self.charset = US_ASCII
    def _read(self, func, size):
        if size == 0:
            return u''
        if size is None or size < 0:
            data = self.data + func()
        else:
            data = self.data + func(max(size, 8) - len(self.data))
        self.data = ''
        if self.charset != US_ASCII:
            data = DESIGNATIONS[self.charset] + data
        pos = data.rfind("\033")
        if pos >= 0 and pos + 3 >= len(data):
            # data ends on the way of an escape sequence
            data, self.data = data[:pos], data[pos:]
            pos = data.rfind("\033")
        if pos >= 0:
            try:
                self.charset = CHARSETS[data[pos:pos+3]]
            except KeyError:
                raise UnicodeError, "unknown designation"
            if self.charset == KSC5601_1987 and \
               (len(data) - pos - 3) % 2 == 1:
                data, self.data = data[:-1], data[-1]
            if self.charset != US_ASCII:
                data = data + DESIGNATIONS[US_ASCII]
        return self.decode(data)[0]
    def read(self, size=-1):
        return self._read(self.stream.read, size)
    def readline(self, size=-1):
        return self._read(self.stream.readline, size)
    def readlines(self, size=-1):
        data = self._read(self.stream.read, size)
        buffer = []
        end = 0
        while 1:
            pos = data.find(u'\n', end)
            if pos < 0:
                if end < len(data):
                    buffer.append(data[end:])
                break
            buffer.append(data[end:pos+1])
            end = pos+1
        return buffer
    def reset(self):
        self.data = ''

### encodings module API

def getregentry():
    return (Codec().encode,Codec().decode,StreamReader,StreamWriter)

def getaliases():
    return ('iso_2022_kr',)



# Hye-Shik Chang <16 Feb 2002>
# originally written by Tamito KAJIYAMA
#
# $Id: euc_kr.py,v 1.1.1.1 2002/02/16 00:51:36 perky Exp $

import codecs

from korean.mappings import ksc5601_hangul
encmap_hangul, decmap_hangul = ksc5601_hangul.encoding_map, ksc5601_hangul.decoding_map
encmap_ideo, decmap_ideo = {}, {}
encmap_misc, decmap_misc = {}, {}

class Codec(codecs.Codec):

    # Unicode to character buffer
    def encode(self, data, errors='strict',
               supported_errors=('strict', 'ignore', 'replace')):
        global encmap_ideo, encmap_misc

        if errors not in supported_errors:
            raise UnicodeError, "unknown error handling"
        buffer = []

        for c in data:
            if c < u'\u0080':
                buffer.append(c.encode("ascii", errors))
            elif encmap_hangul.has_key(c):
                buffer.append(encmap_hangul[c])
            else:
                if not encmap_misc:
                    from korean.mappings import ksc5601_misc
                    encmap_misc = ksc5601_misc.encoding_map
                if encmap_misc.has_key(c):
                    buffer.append(encmap_misc[c])
                    continue

                if not encmap_ideo:
                    from korean.mappings import ksc5601_ideograph
                    encmap_ideo = ksc5601_ideograph.encoding_map
                if encmap_ideo.has_key(c):
                    buffer.append(encmap_ideo[c])
                    continue

                if errors == 'replace':
                    buffer.append('\xa0\xa0')
                elif errors == 'strict':
                    raise UnicodeError, "cannot map \\u%04x to EUC-KR" % ord(c)

        return (''.join(buffer), len(data))

    # character buffer to Unicode
    def decode(self, data, errors='strict',
               supported_errors=('strict', 'ignore', 'replace')):
        global decmap_ideo, decmap_misc

        if errors not in supported_errors:
            raise UnicodeError, "unknown error handling"

        buffer = []
        data = str(data) # character buffer compatible object
        size = len(data)
        p = 0
        while p < size:
            if data[p] < '\x80':
                buffer.append(unicode(data[p], "ascii", errors))
                p += 1
            else:
                c = data[p:p+2]
                if len(c) == 2:
                    if decmap_hangul.has_key(c):
                        buffer.append(decmap_hangul[c])
                        p += 2
                        continue

                    if not decmap_misc:
                        from korean.mappings import ksc5601_misc
                        decmap_misc = ksc5601_misc.decoding_map
                    if decmap_misc.has_key(c):
                        buffer.append(decmap_misc[c])
                        p += 2
                        continue
    
                    if not decmap_ideo:
                        from korean.mappings import ksc5601_ideograph
                        decmap_ideo = ksc5601_ideograph.decoding_map
                    if decmap_ideo.has_key(c):
                        buffer.append(decmap_ideo[c])
                        p += 2
                        continue

                if errors == 'replace':
                    buffer.append(u'\uFFFD') # REPLACEMENT CHARACTER
                elif errors == 'strict':
                    raise UnicodeError, "unexpected byte 0x%02x found" % ord(data[p])
                p += 1

        return (u''.join(buffer), size)


class StreamWriter(Codec, codecs.StreamWriter):
    pass


class StreamReader(Codec, codecs.StreamReader):

    def __init__(self, stream, errors='strict'):
        codecs.StreamReader.__init__(self, stream, errors)
        self.data = ''

    def _read(self, func, size):
        if size == 0:
            return u''
        if size is None or size < 0:
            data = self.data + func()
            self.data = ''
        else:
            data = self.data + func(max(size, 2) - len(self.data))
            size = len(data)
            p = 0
            while p < size:
                if data[p] < "\x80":
                    p = p + 1
                elif p + 2 <= size:
                    p = p + 2
                else:
                    break
            data, self.data = data[:p], data[p:]
        return self.decode(data)[0]

    def read(self, size=-1):
        return self._read(self.stream.read, size)

    def readline(self, size=-1):
        return self._read(self.stream.readline, size)

    def readlines(self, size=-1):
        data = self._read(self.stream.read, size)
        buffer = []
        end = 0
        while 1:
            pos = data.find(u'\n', end)
            if pos < 0:
                if end < len(data):
                    buffer.append(data[end:])
                break
            buffer.append(data[end:pos+1])
            end = pos+1
        return buffer
    def reset(self):
        self.data = ''

### encodings module API

def getregentry():
    return (Codec().encode,Codec().decode,StreamReader,StreamWriter)

def getaliases():
    return ('euc_kr', 'ksc5601', 'ksx1001', 'wansung',)



#!/usr/local/bin/python
# ex:ts=4
#
# Unicode hangul abstractive controller
#
#   written by Hye-Shik Chang <perky@fallin.lv>
#
# Unicode Hangul Code-Area Specifications:
#  http://www.unicode.org/charts/PDF/UAC00.pdf
#
# ---------------------------------------------------------------------------- 
# "THE BEER-WARE LICENSE" (Revision 42): 
# <perky@fallin.lv> wrote this file. As long as you retain this notice you 
# can do whatever you want with this stuff. If we meet some day, and you think 
# this stuff is worth it, you can buy me a beer in return. Hye-Shik Chang
# ---------------------------------------------------------------------------- 
#
# $LinuxKorea: UnicodeHangul.py,v 1.3 2001/08/04 05:51:29 perky Exp $'
# $Id: hangul.py,v 1.1.1.1 2002/02/16 00:51:13 perky Exp $
#

class UnicodeHangulError(Exception):
	
	def __init__ (self, msg):
		self.msg = msg
	
	def __repr__ (self):
		return self.msg
	
	__str__ = __repr__

Null = u''

class Jaeum: # XXX: 1100-1159 Old Jaeum need?

	Codes = (u'\u3131', u'\u3132', u'\u3133', u'\u3134', u'\u3135', u'\u3136',
			#    G         GG          GS         N          NJ         NH
			 u'\u3137', u'\u3138', u'\u3139', u'\u313a', u'\u313b', u'\u313c',
			#    D         DD          R          RG         RM         RB
			 u'\u313d', u'\u313e', u'\u313f', u'\u3140', u'\u3141', u'\u3142',
			#    RS        RT          RP         RH         M          B
			 u'\u3143', u'\u3144', u'\u3145', u'\u3146', u'\u3147', u'\u3148',
			#    BB        BS          S          SS         A          J
			 u'\u3149', u'\u314a', u'\u314b', u'\u314c', u'\u314d', u'\u314e')
			#    JJ        CH          K          T          P          H
	Width = len(Codes)
	G, GG, GS, N, NJ, NH, D, DD, R, RG, RM, RB, RS, RT, RP, RH, M, B, \
	BB, BS, S, SS, A, J, JJ, CH, K, T, P, H = Codes
	Chosung = [G, GG, N, D, DD, R, M, B, BB, S, SS, A, J, JJ, CH, K, T, P, H]
	Jongsung = [Null, G, GG, GS, N, NJ, NH, D, R, RG, RM, RB, RS, RT, \
				 RP, RH, M, B, BS, S, SS, A, J, CH, K, T, P, H]
	MultiElement = {
		GG: (G, G),		GS: (G, S),		NJ: (N, J),		NH: (N, H),		DD: (D, D),
		RG: (R, G),		RM: (R, M),		RB: (R, B),		RS: (R, S),		RT: (R, T),
		RP: (R, P),		RH: (R, H),		BB: (B, B),		BS: (B, S),		SS: (S, S),
		JJ: (J, J)
	}


class Moeum: # XXX: 1161-117f Old Moeum need?

	Codes = (u'\u314f', u'\u3150', u'\u3151', u'\u3152', u'\u3153', u'\u3154',
			#    A          AE        YA         YAE         EO         E
			 u'\u3155', u'\u3156', u'\u3157', u'\u3158', u'\u3159', u'\u315a',
			#    YEO        YE        O          WA          WAE        WOE
			 u'\u315b', u'\u315c', u'\u315d', u'\u315e', u'\u315f', u'\u3160',
			#    YO         OO        WO         WE          WI         YU
			 u'\u3161', u'\u3162', u'\u3163')
			#    EU         EUI       I
	Width = len(Codes)
	A, AE, YA, YAE, EO, E, YEO, YE, O, WA, WAE, WOE, YO, \
	OO, WO, WE, WI, YU, EU, EUI, I = Codes
	Jungsung = list(Codes)
	MultiElement = {
		AE: (A, I),		YAE: (YA, I),	YE: (YEO, I),	WA:	(O, A),		WAE: (O, A, I),
		WOE: (O, I),	WO: (OO, EO),	WE: (OO, E),	WI: (OO, I),	EUI: (EU, I)
	}


# Aliases for your convinience
Chosung = Jaeum.Chosung
Jungsung = Moeum.Jungsung
Jongsung = Jaeum.Jongsung

isJaeum = lambda c: c in Jaeum.Codes
isMoeum = lambda c: c in Moeum.Codes

# Unicode Hangul Syllables Characteristics
zone = (u'\uAC00', u'\uD7A3')
splitters = [ ( len(Jongsung)*len(Jungsung), Chosung ),
			  ( len(Jongsung),				 Jungsung ),
			  ( 1,							 Jongsung ) ]

ishangul = (
	lambda code:
		zone[0] <= code <= zone[1] or
		code in Jaeum.Codes or
		code in Moeum.Codes
)

def join(codes):
	""" Join function which makes hangul syllable from jamos """
	if len(codes) is not 3:
		raise UnicodeHangulError("needs 3-element tuple")
	if not codes[0] or not codes[1]: # single jamo
		return codes[0] or codes[1]

	r = ord(zone[0])
	codes = codes[:]  # simple copy :D
	for multiplier, codeset in splitters:
		r = r + multiplier*codeset.index(codes.pop(0))

	return unichr(r)

def split(code):
	""" Split function which splits hangul syllable into jamos """
	if len(code) != 1 or not ishangul(code):
		raise UnicodeHangulError("needs 1 hangul letter")
	if code in Jaeum.Codes:
		return [code, Null, Null]
	if code in Moeum.Codes:
		return [Null, code, Null]

	code = ord(code) - ord(zone[0])
	r = []
	for divider, codeset in splitters:
		value, code = code / divider, code % divider
		r.append(codeset[value])
	return r

def dividestring(str, intoelements=0):
	if type(str) is not type(u''):
		raise UnicodeHangulError("needs unicode string")

	r = u''
	for char in str:
		if ishangul(char):
			elems = split(char)
			for elem in elems:
				for htype in (Jaeum, Moeum, None):
					if htype == None:
						r += elem
					elif intoelements and \
					   htype.MultiElement.has_key(elem):
						r += u''.join(htype.MultiElement[elem])
						break
		else:
			r += char
	
	return r


if __name__ == '__main__':

	print ( join([Jaeum.P, Moeum.EO, Null]) + \
			join([Jaeum.K, Moeum.I, Null]) + \
			join([Jaeum.JJ, Moeum.A, Jaeum.A]) ).encode("utf-8")
	
	while 1:
		code = raw_input(">>> ")
		print dividestring(unicode(code, "utf-8"), 1).encode("utf-8")




# Hye-Shik Chang <16 Feb 2002>
# $Id: qwerty2bul.py,v 1.1.1.1 2002/02/16 00:51:13 perky Exp $

try:
    from korean.c.qwerty2bul import *
except ImportError:
    from korean.python.qwerty2bul import *



# Hye-Shik Chang <16 Feb 2002>
# $Id: johab.py,v 1.1.1.1 2002/02/16 00:51:36 perky Exp $

import codecs

from korean.hangul import Jaeum, Moeum, ishangul, split, join
encmap, decmap = {}, {}

johab2uni_chosung = {
    1: u'',         2: Jaeum.G,     3: Jaeum.GG,    4: Jaeum.N,
    5: Jaeum.D,     6: Jaeum.DD,    7: Jaeum.R,     8: Jaeum.M,
    9: Jaeum.B,     10: Jaeum.BB,   11: Jaeum.S,    12: Jaeum.SS,
    13: Jaeum.A,    14: Jaeum.J,    15: Jaeum.JJ,   16: Jaeum.CH,
    17: Jaeum.K,    18: Jaeum.T,    19: Jaeum.P,    20: Jaeum.H,
}
johab2uni_jungsung = {
    2: u'',         3: Moeum.A,     4: Moeum.AE,    5: Moeum.YA,
    6: Moeum.YAE,   7: Moeum.EO,    10: Moeum.E,    11: Moeum.YEO,
    12: Moeum.YE,   13: Moeum.O,    14: Moeum.WA,   15: Moeum.WAE,
    18: Moeum.WOE,  19: Moeum.YO,   20: Moeum.OO,   21: Moeum.WO,
    22: Moeum.WE,   23: Moeum.WI,   26: Moeum.YU,   27: Moeum.EU,
    28: Moeum.EUI,  29: Moeum.I
}
johab2uni_jongsung = {
    1: u'',         2: Jaeum.G,     3: Jaeum.GG,    4: Jaeum.GS,
    5: Jaeum.N,     6: Jaeum.NJ,    7: Jaeum.NH,    8: Jaeum.D,
    9: Jaeum.R,     10: Jaeum.RG,   11: Jaeum.RM,   12: Jaeum.RB,
    13: Jaeum.RS,   14: Jaeum.RT,   15: Jaeum.RP,   16: Jaeum.RH,
    17: Jaeum.M,    19: Jaeum.B,    20: Jaeum.BS,   21: Jaeum.S,
    22: Jaeum.SS,   23: Jaeum.A,    24: Jaeum.J,    25: Jaeum.CH,
    26: Jaeum.K,    27: Jaeum.T,    28: Jaeum.P,    29: Jaeum.H
}

uni2johab_chosung = {}
uni2johab_jungsung = {}
uni2johab_jongsung = {}
for k, v in johab2uni_chosung.items():
    uni2johab_chosung[v] = k
for k, v in johab2uni_jungsung.items():
    uni2johab_jungsung[v] = k
for k, v in johab2uni_jongsung.items():
    uni2johab_jongsung[v] = k


class Codec(codecs.Codec):

    # Unicode to character buffer
    def encode(self, data, errors='strict',
               supported_errors=('strict', 'ignore', 'replace')):
        global encmap

        if errors not in supported_errors:
            raise UnicodeError, "unknown error handling"
        buffer = []

        for c in data:
            if c < u'\u0080':
                buffer.append(c.encode("ascii", errors))
            elif ishangul(c):
                cho, jung, jong = split(c) # all hangul can success
                cho, jung, jong = (
                    uni2johab_chosung[cho],
                    uni2johab_jungsung[jung],
                    uni2johab_jongsung[jong]
                )
                code = 0x8080 | (cho<<10) | (jung<<5) | jong
                buffer.append(chr(code>>8) + chr(code&0xFF))
            else:
                if not encmap:
                    from korean.mappings import johab_ideograph
                    encmap = johab_ideograph.encoding_map

                if encmap.has_key(c):
                    buffer.append(encmap[c])
                elif errors == 'replace':
                    buffer.append('\x84\x41')
                elif errors == 'strict':
                    raise UnicodeError, "cannot map \\u%04x to JOHAB" % ord(c)

        return (''.join(buffer), len(data))

    # character buffer to Unicode
    def decode(self, data, errors='strict',
               supported_errors=('strict', 'ignore', 'replace')):
        global decmap

        if errors not in supported_errors:
            raise UnicodeError, "unknown error handling"

        buffer = []
        data = str(data) # character buffer compatible object
        size = len(data)
        p = 0
        while p < size:
            if data[p] < '\x80':
                buffer.append(unicode(data[p], "ascii", errors))
                p += 1
            else:
                c = data[p:p+2]
                if len(c) == 2:
                    code = (ord(c[0])<<8) | ord(c[1])
                    cho = (code >> 10) & 0x1f
                    jung = (code >> 5) & 0x1f
                    jong = (code) & 0x1f
                    if ( johab2uni_chosung.has_key(cho) and
                         johab2uni_jungsung.has_key(jung) and
                         johab2uni_jongsung.has_key(jong) ):
                        buffer.append( join([
                            johab2uni_chosung[cho],
                            johab2uni_jungsung[jung],
                            johab2uni_jongsung[jong]
                        ]) )
                        p += 2
                        continue
                        
                    if not decmap:
                        from korean.mappings import johab_ideograph
                        decmap = johab_ideograph.decoding_map

                    if decmap.has_key(c):
                        buffer.append(decmap[c])
                        p += 2
                        continue

                if errors == 'replace':
                    buffer.append(u'\uFFFD') # REPLACEMENT CHARACTER
                elif errors == 'strict':
                    raise UnicodeError, "unexpected byte 0x%02x found" % ord(data[p])
                p += 1

        return (u''.join(buffer), size)


class StreamWriter(Codec, codecs.StreamWriter):
    pass


class StreamReader(Codec, codecs.StreamReader):

    def __init__(self, stream, errors='strict'):
        codecs.StreamReader.__init__(self, stream, errors)
        self.data = ''

    def _read(self, func, size):
        if size == 0:
            return u''
        if size is None or size < 0:
            data = self.data + func()
            self.data = ''
        else:
            data = self.data + func(max(size, 2) - len(self.data))
            size = len(data)
            p = 0
            while p < size:
                if data[p] < "\x80":
                    p = p + 1
                elif p + 2 <= size:
                    p = p + 2
                else:
                    break
            data, self.data = data[:p], data[p:]
        return self.decode(data)[0]

    def read(self, size=-1):
        return self._read(self.stream.read, size)

    def readline(self, size=-1):
        return self._read(self.stream.readline, size)

    def readlines(self, size=-1):
        data = self._read(self.stream.read, size)
        buffer = []
        end = 0
        while 1:
            pos = data.find(u'\n', end)
            if pos < 0:
                if end < len(data):
                    buffer.append(data[end:])
                break
            buffer.append(data[end:pos+1])
            end = pos+1
        return buffer
    def reset(self):
        self.data = ''

### encodings module API

def getregentry():
    return (Codec().encode,Codec().decode,StreamReader,StreamWriter)

def getaliases():
    return ('johab',)



# Hye-Shik Chang <16 Feb 2002>
# $Id: __init__.py,v 1.2 2002/02/17 13:48:00 perky Exp $

__aliases__ = {
    'korean.cp949':         ('cp949', 'ms949', '949', 'uhc'),
    'korean.euc_kr':        ('euc_kr', 'ksc5601', 'ksc5601_1987'),
    'korean.iso_2022_kr':   ('iso_2022_kr',),
    'korean.johab':         ('johab',),
    'korean.qwerty2bul':    ('qwerty2bul', '2bul'),
    'korean.unijohab':      ('unijohab', 'unicode_johab', 'macjohab'),
}

from encodings.aliases import aliases

for enc, als in __aliases__.items():
    for aliasenam in als:
        aliases[aliasenam] = enc

del aliases, als, enc, aliasenam



# Hye-Shik Chang <16 Feb 2002>
# originally written by Tamito KAJIYAMA
#
# $Id: cp949.py,v 1.1.1.1 2002/02/16 00:51:35 perky Exp $

from korean.python.euc_kr import *

from korean.mappings import uhc
encmap_hangul.update(uhc.encoding_map)
decmap_hangul.update(uhc.decoding_map)
del uhc

def getaliases():
    return ('cp949', 'uhc', 'ms949',)



# Hye-Shik Chang <16 Feb 2002>
# $Id: unijohab.py,v 1.1 2002/02/16 11:34:25 perky Exp $

import codecs

from korean.hangul import Jaeum, Moeum, ishangul, split, join
encmap, decmap = {}, {}

johab2uni_chosung = {
    u'\u115f': u'',         u'\u1100': Jaeum.G,     u'\u1101': Jaeum.GG,
    u'\u1102': Jaeum.N,     u'\u1103': Jaeum.D,     u'\u1104': Jaeum.DD,
    u'\u1105': Jaeum.R,     u'\u1106': Jaeum.M,     u'\u1107': Jaeum.B,
    u'\u1108': Jaeum.BB,    u'\u1109': Jaeum.S,     u'\u110a': Jaeum.SS,
    u'\u110b': Jaeum.A,     u'\u110c': Jaeum.J,     u'\u110d': Jaeum.JJ,
    u'\u110e': Jaeum.CH,    u'\u110f': Jaeum.K,     u'\u1110': Jaeum.T,
    u'\u1111': Jaeum.P,     u'\u1112': Jaeum.H
}
johab2uni_jungsung = {
    u'\u1160': u'',         u'\u1161': Moeum.A,     u'\u1162': Moeum.AE,
    u'\u1163': Moeum.YA,    u'\u1164': Moeum.YAE,   u'\u1165': Moeum.EO,
    u'\u1166': Moeum.E,     u'\u1167': Moeum.YEO,   u'\u1168': Moeum.YE,
    u'\u1169': Moeum.O,     u'\u116a': Moeum.WA,    u'\u116b': Moeum.WAE,
    u'\u116c': Moeum.WOE,   u'\u116d': Moeum.YO,    u'\u116e': Moeum.OO,
    u'\u116f': Moeum.WO,    u'\u1170': Moeum.WE,    u'\u1171': Moeum.WI,
    u'\u1172': Moeum.YU,    u'\u1173': Moeum.EU,    u'\u1174': Moeum.EUI,
    u'\u1175': Moeum.I
}
johab2uni_jongsung = {
    u'': u'',               u'\u11a8': Jaeum.G,     u'\u11a9': Jaeum.GG,
    u'\u11aa': Jaeum.GS,    u'\u11ab': Jaeum.N,     u'\u11ac': Jaeum.NJ,
    u'\u11ad': Jaeum.NH,    u'\u11ae': Jaeum.D,     u'\u11af': Jaeum.R,
    u'\u11b0': Jaeum.RG,    u'\u11b1': Jaeum.RM,    u'\u11b2': Jaeum.RB,
    u'\u11b3': Jaeum.RS,    u'\u11b4': Jaeum.RT,    u'\u11b5': Jaeum.RP,
    u'\u11b6': Jaeum.RH,    u'\u11b7': Jaeum.M,     u'\u11b8': Jaeum.B,
    u'\u11b9': Jaeum.BS,    u'\u11ba': Jaeum.S,     u'\u11bb': Jaeum.SS,
    u'\u11bc': Jaeum.A,     u'\u11bd': Jaeum.J,     u'\u11be': Jaeum.CH,
    u'\u11bf': Jaeum.K,     u'\u11c0': Jaeum.T,     u'\u11c1': Jaeum.P,
    u'\u11c2': Jaeum.H
}

uni2johab_chosung = {}
uni2johab_jungsung = {}
uni2johab_jongsung = {}
for k, v in johab2uni_chosung.items():
    uni2johab_chosung[v] = k
for k, v in johab2uni_jungsung.items():
    uni2johab_jungsung[v] = k
for k, v in johab2uni_jongsung.items():
    uni2johab_jongsung[v] = k


class Codec(codecs.Codec):

    # Unicode to character buffer
    def encode(self, data, errors='strict',
               supported_errors=('strict', 'ignore', 'replace')):
        global encmap

        if errors not in supported_errors:
            raise UnicodeError, "unknown error handling"
        buffer = []

        for c in data:
            if ishangul(c):
                cho, jung, jong = split(c) # all hangul can success
                buffer.append(
                   uni2johab_chosung[cho] +
                    uni2johab_jungsung[jung] +
                    uni2johab_jongsung[jong]
                )
            else:
                buffer.append(c)

        return (u''.join(buffer).encode('utf-8', errors), len(data))

    # character buffer to Unicode
    def decode(self, data, errors='strict',
               supported_errors=('strict', 'ignore', 'replace')):
        global decmap

        if errors not in supported_errors:
            raise UnicodeError, "unknown error handling"

        buffer = []
        data = unicode(data, 'utf-8', errors)
        size = len(data)
        p = 0
        while p < size:
            if not u'\u1100' <= data[p] <= u'\u11FF':
                buffer.append(data[p])
                p += 1
            else:
                c = data[p:p+3]
                try:
                    cho = johab2uni_chosung[c[0]]
                    jung = johab2uni_jungsung[c[1]]
                    if len(c)>2 and johab2uni_jongsung.has_key(c[2]):
                        jong = johab2uni_jongsung[c[2]]
                        p += 3 # this must locate end of this block
                    else:
                        jong = u''
                        p += 2 # too.
                except:
                    if errors == 'replace':
                        buffer.append(u'\uFFFD') # REPLACEMENT CHARACTER
                    elif errors == 'strict':
                        raise UnicodeError, "unexpected byte \\u%04x found" % ord(c[0])
                    p += 1
                else:
                    buffer.append(join([cho, jung, jong]))

        return (u''.join(buffer), size)


class StreamWriter(Codec, codecs.StreamWriter):
    pass

class StreamReader(Codec, codecs.StreamReader):
    pass
    # XXX: Temporarily None.

### encodings module API

def getregentry():
    return (Codec().encode,Codec().decode,StreamReader,StreamWriter)

def getaliases():
    return ('unijohab',)
